from pathlib import Path


class PathManager:
    """Utility for managing filesystem paths via DataInterface and ConfigInterface."""

    def __init__(self, **paths: Path | str) -> None:
        """Initialize key-value pairs for paths."""
        self.paths: dict[str, Path] = {}
        for key, path in paths.items():
            self.add_path(key, Path(path))

    def add_path(
        self, key: str, path: Path | str, exist_ok: bool = False
    ) -> None:
        if not exist_ok and key in self.paths:
            raise ValueError(f"{key} already exists")
        self.paths[key] = Path(path)

    def get_path(self, key: str) -> Path:
        if key not in self.paths:
            raise ValueError(f"Path for '{key}' not found.")
        return self.paths[key]

    def remove_path(self, key: str) -> None:
        if key not in self.paths:
            raise ValueError(f"Path for '{key}' not found.")
        del self.paths[key]

    def get_full_path(self, *fparts: str, key: str | None = None) -> Path:
        """Retrieve the full path based on key and sub-paths."""
        base_dir = Path("") if key is None else self.get_path(key)
        return base_dir / "/".join(map(str, fparts))

    def __repr__(self) -> str:
        expr = f"{type(self).__name__}(\n"
        for key, path in self.paths.items():
            expr += f"    {key}={path},\n"
        expr += ")"
        return expr



from pydantic import BaseModel, model_validator
from pydantic import FilePath as PydanticFilePath


class FilePath(BaseModel):
    path: PydanticFilePath
    extension: str | None = None

    @model_validator(mode="after")
    def check_extension(self):
        path = self.path
        extension = self.extension
        if extension is not None and path.suffix != extension:
            raise ValueError(
                f"File extension '{path.suffix}' does not match expected extension '{extension}'"
            )
        return self



"""Configuration classes."""

from typing import Any

from pydantic import BaseModel, ConfigDict

from onemod.dtypes.unique_sequence import unique_list


class Config(BaseModel):
    """Base configuration class.

    Config instances are dictionary-like objects that contains settings.
    For attribute validation, users can create custom configuration
    classes by subclassing Config. Alternatively, users can add extra
    attributes to Config instances without validation.

    """

    model_config = ConfigDict(
        extra="allow", validate_assignment=True, protected_namespaces=()
    )

    def get(self, key: str, default: Any = None) -> Any:
        if self.__contains__(key):
            return getattr(self, key)
        return default

    def __getitem__(self, key: str) -> Any:
        if self.__contains__(key):
            return getattr(self, key)
        raise KeyError(f"Invalid config item: {key}")

    def __setitem__(self, key: str, value: Any) -> None:
        setattr(self, key, value)

    def __contains__(self, key: str) -> bool:
        return key in self._get_fields()

    def __repr__(self) -> str:
        arg_list = []
        for key in self._get_fields():
            arg_list.append(f"{key}={getattr(self, key)!r}")
        return f"{type(self).__name__}({', '.join(arg_list)})"

    def _get_fields(self) -> list[str]:
        return list(self.model_dump(exclude_none=True))


class StageConfig(Config):
    """Stage configuration class.

    If a StageConfig instance does not contain an attribute, the get and
    __getitem__ methods will return the corresponding pipeline
    attribute, if it exists.

    """

    model_config = ConfigDict(
        extra="allow", validate_assignment=True, protected_namespaces=()
    )

    _pipeline_config: Config = Config()
    _required: list[str] = []

    def add_pipeline_config(self, pipeline_config: Config | dict) -> None:
        if isinstance(pipeline_config, dict):
            pipeline_config = Config(**pipeline_config)

        missing = []
        for item in self._required:
            if not self.stage_contains(item) and item not in pipeline_config:
                missing.append(item)
        if missing:
            raise AttributeError(f"Missing required config items: {missing}")

        self._pipeline_config = pipeline_config

    def get(self, key: str, default: Any = None) -> Any:
        if self.stage_contains(key):
            return getattr(self, key)
        return self._pipeline_config.get(key, default)

    def get_from_stage(self, key: str, default: Any = None) -> Any:
        if self.stage_contains(key):
            return getattr(self, key)
        return default

    def get_from_pipeline(self, key: str, default: Any = None) -> Any:
        return self._pipeline_config.get(key, default)

    def __getitem__(self, key: str) -> Any:
        if self.stage_contains(key):
            return getattr(self, key)
        return self._pipeline_config[key]

    def __contains__(self, key: str) -> bool:
        return self.stage_contains(key) or self.pipeline_contains(key)

    def stage_contains(self, key: str) -> bool:
        return key in self._get_stage_fields()

    def pipeline_contains(self, key: str) -> bool:
        return key in self._pipeline_config

    def _get_fields(self) -> list[str]:
        return unique_list(
            self._get_stage_fields() + self._get_pipeline_fields()
        )

    def _get_stage_fields(self) -> list[str]:
        return list(self.model_dump(exclude_none=True))

    def _get_pipeline_fields(self) -> list[str]:
        return self._pipeline_config._get_fields()

    def __repr__(self) -> str:
        arg_list = []
        for key in self._get_fields():
            arg_list.append(f"{key}={self.get(key)!r}")
        return f"{type(self).__name__}({', '.join(arg_list)})"



"""Methods to load and evaluate pipeline and stage objects."""

import json
from importlib.util import module_from_spec, spec_from_file_location
from inspect import getmodulename
from pathlib import Path
from typing import Any, Literal

import fire

import onemod.stage as onemod_stages
from onemod.pipeline import Pipeline
from onemod.stage import Stage


def load_pipeline(config: Path | str) -> Pipeline:
    """Load pipeline instance from JSON file.

    Parameters
    ----------
    config : Path or str
        Path to pipeline config file.

    Returns
    -------
    Pipeline
        Pipeline instance.

    """
    return Pipeline.from_json(config)


def load_stage(config: Path | str, stage_name: str) -> Stage:
    """Load stage instance from JSON file.

    Parameters
    ----------
    config : Path or str
        Path to pipeline config file.
    stage_name : str
        Stage name.

    Returns
    -------
    Stage
        Stage instance.

    """
    stage_class = _get_stage(config, stage_name)
    stage = stage_class.from_json(config, stage_name)
    return stage


def _get_stage(config: Path | str, stage_name: str) -> Stage:
    """Get stage class from JSON file.

    Parameters
    ----------
    config : Path or str
        Path to config file.
    stage_name : str
        Stage name.

    Returns
    -------
    Stage
        Stage class.

    Notes
    -----
    When a custom stage class has the same name as a built-in OneMod
    stage class, this function returns the custom stage class.

    """
    with open(config, "r") as f:
        config_dict = json.load(f)
    if stage_name not in config_dict["stages"]:
        raise KeyError(f"Config does not contain a stage named '{stage_name}'")
    config_dict = config_dict["stages"][stage_name]
    stage_type = config_dict["type"]

    if "module" in config_dict:
        return _get_custom_stage(stage_type, config_dict["module"])
    if hasattr(onemod_stages, stage_type):
        return getattr(onemod_stages, stage_type)
    raise KeyError(
        f"Config does not contain a module for custom stage '{stage_name}'"
    )


def _get_custom_stage(stage_type: str, module: str) -> Stage:
    """Get custom stage class from file.

    Parameters
    ----------
    stage_type : str
        Name of custom stage class.
    module : str
        Path to Python module containing custom stage class definition.

    Returns
    -------
    Stage
        Custom stage class.

    """
    module_path = Path(module)

    module_name = getmodulename(module_path)
    if module_name is None:
        raise ValueError(f"Could not determine module name from {module_path}")

    spec = spec_from_file_location(module_name, module_path)
    if spec is None:
        raise ImportError(f"Could not load spec for module {module_path}")

    if spec.loader is None:
        raise ImportError(f"Module spec for {module_path} has no loader")

    loaded_module = module_from_spec(spec)
    spec.loader.exec_module(loaded_module)

    return getattr(loaded_module, stage_type)


def evaluate(
    config: Path | str,
    method: Literal["run", "fit", "predict", "collect"] = "run",
    stages: str | list[str] | None = None,
    subsets: dict[str, Any | list[Any]] | None = None,
    paramsets: dict[str, Any | list[Any]] | None = None,
    collect: bool | None = None,
    backend: Literal["local", "jobmon"] = "local",
    cluster: str | None = None,
    resources: Path | str | dict[str, Any] | None = None,
    python: Path | str | None = None,
    **kwargs,
) -> None:
    """Evaluate pipeline or stage method.

    When evaluating a pipeline method, all stage submodels are evaluated
    and submodel results are collected.

    Parameters
    ----------
    config : Path or str
        Path to pipeline config file.
    method : str, optional
        Name of method to evaluate. Default is 'run'.
    stages : str, list of str, optional
        Names of stages to evaluate. If None, evaluate all pipeline
        stages. Default is None.
    backend : str, optional
        Whether to evaluate the method locally or with Jobmon.
        Default is 'local'.
    **kwargs
        Additional keyword arguments passed to stage methods. When
        evaluating a pipeline, use format ``stage={arg_name: arg_value}``.

    Stage Parameters
    ----------------
    subsets : dict, optional
        Submodel data subsets to include when evaluating a single stage.
        If None, evaluate all data subsets. Default is None.
    paramsets : dict, optional
        Submodel parameter sets to include when evaluating a single
        stage. If None, evaluate all parameter sets. Default is None.
    collect : bool, optional
        Whether to collect submodel results when evaluating a single
        stage. If ``subsets`` and ``paramsets`` are both None, default
        is True, otherwise default is False.

    Jobmon Parameters
    -----------------
    cluster : str, optional
        Cluster name. Required if ``backend`` is 'jobmon'.
    resources : Path, str, or dict, optional
        Path to resources file or dictionary of compute resources.
        Required if ``backend`` is 'jobmon'.
    python : Path or str, optional
        Path to Python environment if ``backend`` is 'jobmon'. If None,
        use sys.executable. Default is None.

    Examples
    --------
    This method is the entrypoint when calling onemod from the command
    line:

    .. code:: bash

       onemod --config pipeline_config.json --method run

    To make sure arguments such as lists or dictionaries are parsed
    correctly, omit spaces or use quotes:

    .. code:: bash

       onemod --config pipeline_config.json --stages [stage1,stage2]
       onemod --config pipeline_config.json --stages stage1 --subsets "{'age_group_id': 1}"

    See the `Python Fire Guide <https://google.github.io/python-fire/guide/#argument-parsing>`_
    for more details.

    """
    model: Pipeline | Stage

    if isinstance(stages, str):
        model = load_stage(config, stages)
        model.evaluate(
            method,
            subsets,
            paramsets,
            collect,
            backend,
            cluster,
            resources,
            python,
            **kwargs,
        )
    else:
        model = load_pipeline(config)
        model.evaluate(
            method, stages, backend, cluster, resources, python, **kwargs
        )


def main():
    fire.Fire(evaluate)


if __name__ == "__main__":
    main()



from typing import Any, Callable

from polars import Series
from pydantic import BaseModel, Field, field_validator

from onemod.constraints.functions import bounds, is_in

# Global registry for constraints
CONSTRAINT_REGISTRY: dict[str, Callable] = {}


class Constraint(BaseModel):
    name: str
    args: dict[str, Any]

    func: Callable[[Series], None] | None = Field(default=None, exclude=True)

    def model_post_init(self, __context):
        """Reconstruct the `func` attribute after deserialization."""
        self.func = CONSTRAINT_REGISTRY[self.name](**self.args)

    @field_validator("name")
    def validate_name(cls, value):
        """Ensure the constraint name is in the global registry."""
        if value not in CONSTRAINT_REGISTRY:
            raise ValueError(
                f"Unknown constraint: {value}. Did you forget to register it?"
            )
        return value

    def use_validation(self, column: Series) -> None:
        """Applies the constraint's validation function to a Polars Series."""
        if self.func is None:
            self.func = CONSTRAINT_REGISTRY[self.name](**self.args)
        self.func(column)

    def to_dict(self) -> dict:
        """Convert the constraint to a dictionary for serialization."""
        return {"name": self.name, "args": self.args}

    @classmethod
    def from_dict(cls, data: dict) -> "Constraint":
        """Reconstruct a Constraint from a dictionary."""
        name = data["name"]
        args = data["args"]
        return cls(name=name, args=args)


def register_constraint(name: str, func: Callable) -> None:
    """
    Allows users to register custom constraint functions.

    Parameters
    ----------
    name : str
        The name of the constraint.
    func : callable
        The validation function to apply.

    Examples
    --------
    >>> def custom_constraint_example(limit: int) -> Callable:
    ...     def use_validation(column: Series) -> None:
    ...         if not column.lt(limit).all():
    ...             raise ValueError(f"Values must be less than {limit}.")
    ...
    ...     return use_validation
    >>> register_constraint("custom_constraint", custom_constraint_example)
    """
    if name in CONSTRAINT_REGISTRY:
        raise ValueError(f"Constraint '{name}' is already registered.")
    CONSTRAINT_REGISTRY[name] = func


def register_preset_constraints() -> None:
    """Registers preset constraint functions."""
    register_constraint("bounds", bounds)
    register_constraint("is_in", is_in)


register_preset_constraints()



"""Stage base classes."""

from __future__ import annotations

import json
from abc import ABC
from inspect import getfile
from itertools import product
from pathlib import Path
from typing import Any, Literal, Mapping

from pandas import DataFrame
from pydantic import BaseModel, ConfigDict

import onemod.stage as onemod_stages
from onemod.config import StageConfig
from onemod.dtypes import Data
from onemod.dtypes.unique_sequence import UniqueList
from onemod.fsutils import DataInterface
from onemod.io import Input, Output
from onemod.utils.decorators import computed_property
from onemod.validation import ValidationErrorCollector, handle_error


class Stage(BaseModel, ABC):
    """Stage base class.

    Parameters
    ----------
    name : str
        Stage name.
    config : StageConfig, optional
        Stage configuration.
    groupby : list of str, optional
        Column names used to create submodel data subsets.
    crossby : list of str, optional
        Parameter names used to create submodel parameter sets.
    input_validation : dict of str: Data, optional
        Optional specification of input data validation.
    output_validation : dict of str: Data, optional
        Optional specification of output data validation.

    """

    model_config = ConfigDict(validate_assignment=True)

    name: str
    config: StageConfig = StageConfig()
    groupby: UniqueList[str] | None = None
    crossby: UniqueList[str] | None = None
    input_validation: dict[str, Data] | None = None
    output_validation: dict[str, Data] | None = None
    _module: Path | None = None
    _input: Input
    _output: Output
    _dataif: DataInterface
    _subsets: DataFrame | None = None
    _paramsets: DataFrame | None = None
    _skip: list[str] = []
    _collect_after: list[str] = []
    _required_input: dict[str, dict[str, Any]] = {}
    _optional_input: dict[str, dict[str, Any]] = {}
    _output_items: dict[str, dict[str, Any]] = {}

    def __init__(
        self,
        module: Path | str | None = None,
        input: Input | dict = {},
        config_path: Path | str | None = None,
        **kwargs,
    ) -> None:
        """Create stage instance."""
        super().__init__(**kwargs)
        self.set_module(module)
        self.set_input(input)
        if config_path is not None:
            self.set_dataif(config_path)
            self.set_output()

    @computed_property
    def type(self) -> str:
        """Stage type."""
        return type(self).__name__

    @computed_property
    def module(self) -> Path | None:
        """Path to module containing custom stage definition."""
        return self._module

    def set_module(self, module: Path | str | None) -> None:
        if isinstance(module, (Path, str)):
            self._module = Path(module)
        else:
            if not hasattr(onemod_stages, self.type):
                try:
                    self._module = Path(getfile(self.__class__))
                except TypeError:
                    raise TypeError(
                        f"Could not find module for custom stage '{self.name}'"
                    )

    @computed_property
    def input(self) -> Input:
        """Stage input metadata."""
        # TODO: Could add more description, like keys, Data
        return self._input

    def set_input(self, input: Data | dict) -> None:
        self._input = Input(
            stage=self.name,
            required=self._required_input,
            optional=self._optional_input,
            items=input,
        )

    @property
    def output(self) -> Output:
        """Stage output metadata."""
        # TODO: Could add more description, like keys, Data
        return self._output

    def set_output(self) -> None:
        self._output = Output(
            stage=self.name,
            directory=self.dataif.get_path("output"),
            items=self._output_items,
        )

    @property
    def dependencies(self) -> list[str]:
        """Stage dependencies."""
        return self.input.dependencies

    @property
    def dataif(self) -> DataInterface:
        """Stage data interface."""
        # TODO: Add more detailed description and examples
        return self._dataif

    def set_dataif(self, config_path: Path | str) -> None:
        # Create data interface
        directory = Path(config_path).parent
        self._dataif = DataInterface(
            config=config_path,
            directory=directory,
            output=directory / self.name,
        )

        # Add input items
        for item_name, item_value in self.input.items.items():
            if item_value.path is not None:
                self._dataif.add_path(item_name, item_value.path)

    @property
    def skip(self) -> list[str]:
        """Names of methods skipped by the stage."""
        return self._skip

    @property
    def subsets(self) -> DataFrame | None:
        """Stage data subsets."""
        if self.groupby is not None and self._subsets is None:
            try:
                self._subsets = self.dataif.load("subsets.csv", key="output")
            except FileNotFoundError:
                raise AttributeError(
                    f"Stage '{self.name}' submodel data subsets have not been created"
                )
        return self._subsets

    def create_subsets(self, groupby_data: Path | str) -> None:
        """Create submodel data subsets from groupby."""
        if self.groupby is None:
            raise AttributeError(
                f"Stage '{self.name}' does not use groupby attribute"
            )

        data = self.dataif.load(
            str(groupby_data), columns=self.groupby
        ).drop_duplicates()
        groups = data.groupby(self.groupby)
        self._subsets = DataFrame(
            [subset for subset in groups.groups.keys()], columns=self.groupby
        ).sort_values(by=self.groupby)
        self.dataif.dump(self._subsets, "subsets.csv", key="output")

    @staticmethod
    def get_subset(
        data: DataFrame, subset: Mapping[str, Any | list[Any]]
    ) -> DataFrame:
        """Filter data by subset."""
        for col, values in subset.items():
            data = data[
                data[col].isin(values if isinstance(values, list) else [values])
            ]

        if len(data) == 0:
            raise ValueError(f"Empty subset or paramset: {subset}")

        return data.reset_index(drop=True)

    @property
    def paramsets(self) -> DataFrame | None:
        """Stage parameter sets."""
        if self.crossby is not None and self._paramsets is None:
            try:
                self._paramsets = self.dataif.load(
                    "paramsets.csv", key="output"
                )
            except FileNotFoundError:
                raise AttributeError(
                    f"Stage '{self.name}' submodel parameter sets have not been created"
                )
        return self._paramsets

    def create_params(self) -> None:
        """Create submodel parameter sets from crossby."""
        if self.crossby is None:
            raise AttributeError(
                f"Stage '{self.name}' does not use crossby attribute"
            )

        # Get all parameters with multiples values from config
        param_dict = {}
        for param_name in self.crossby:
            param_values = self.config[param_name]
            if isinstance(param_values, (list, set, tuple)):
                param_dict[param_name] = param_values
            else:
                raise ValueError(
                    f"Crossby param '{param_name}' must be a list, set, or tuple"
                )

        # Create parameter sets
        self._paramsets = DataFrame(
            list(product(*param_dict.values())), columns=list(param_dict.keys())
        ).sort_values(by=self.crossby)
        self.dataif.dump(self._paramsets, "paramsets.csv", key="output")

    def set_params(self, paramset: dict[str, Any]) -> None:
        """Set submodel parameters."""
        if self.crossby is None:
            raise AttributeError(
                f"Stage '{self.name}' does not use crossby attribute"
            )

        for param_name in self.crossby:
            if param_name not in paramset:
                raise KeyError(
                    f"Stage '{self.name}' param set missing param: {param_name}"
                )
            self.config[param_name] = paramset[param_name]

    @property
    def has_submodels(self) -> bool:
        """Whether the stage has submodels."""
        return self.groupby is not None or self.crossby is not None

    def get_submodels(
        self,
        subsets: dict[str, Any | list[Any]] | None = None,
        paramsets: dict[str, Any | list[Any]] | None = None,
    ) -> list[tuple[dict[str, Any] | None, ...]]:
        """Get stage submodels.

        A stage submodel consists of a single ``subset`` / ``paramset``
        combination.

        Parameters
        ----------
        subsets : dict, optional
            Submodel data subsets to include. If None, include all
            submodel data subsets. Default is None.
        paramsets : dict, optional
            Submodel parameter sets to include. If None, include all
            submodel parameter sets. Default is None.

        Returns
        -------
        list of tuple
            Stage submodels.

        """
        # TODO: Add examples?
        if not self.has_submodels:
            raise AttributeError(f"Stage '{self.name}' does not have submodels")
        if subsets is not None and self.subsets is None:
            raise AttributeError(
                f"Stage '{self.name}' does not use groupby attribute"
            )
        if paramsets is not None and self.paramsets is None:
            raise AttributeError(
                f"Stage '{self.name}' does not use crossby attribute"
            )

        # Filter data subsets and parameter sets
        filtered_subsets = self.subsets
        if filtered_subsets is not None and subsets is not None:
            filtered_subsets = self.get_subset(filtered_subsets, subsets)
        filtered_paramsets = self.paramsets
        if filtered_paramsets is not None and paramsets is not None:
            filtered_paramsets = self.get_subset(filtered_paramsets, paramsets)

        # Generate all data subset/parameter set combinations
        return list(
            product(
                [None]
                if filtered_subsets is None
                else filtered_subsets.to_dict(orient="records"),
                [None]
                if filtered_paramsets is None
                else filtered_paramsets.to_dict(orient="records"),
            )  # type: ignore
        )

    @property
    def collect_after(self) -> list[str]:
        """Names of methods that collect submodel results."""
        return self._collect_after

    def get_field(
        self, field: str, stage_name: str | None = None, default: Any = None
    ) -> Any:
        """Get field from config file.

        Parameters
        ----------
        field : str
            Name of field. If field is nested, join keys with ':'.
            For example, 'config:param`.
        stage_name : str or None, optional
            Name of stage if field belongs to stage. Default is None.

        Returns
        -------
        Any
            Field item.

        """
        config = self.dataif.load(key="config")
        if stage_name is not None:
            config = config.get("stages", {}).get(stage_name, {})
        for key in field.split(":"):
            config = config.get(key, {})
        return config or default

    @classmethod
    def from_json(cls, config_path: Path | str, stage_name: str) -> Stage:
        """Load stage from JSON file.

        Parameters
        ----------
        config_path : Path or str
            Path to pipeline config file.
        stage_name : str
            Stage name.

        Returns
        -------
        Stage
            Stage instance.

        """
        with open(config_path, "r") as file:
            pipeline_config = json.load(file)
        try:
            stage_config = pipeline_config["stages"][stage_name]
            del stage_config["type"]
        except KeyError:
            raise AttributeError(
                f"{pipeline_config['name']} does not contain a stage named '{stage_name}'"
            )

        stage = cls(config_path=config_path, **stage_config)
        stage.config.add_pipeline_config(pipeline_config["config"])
        return stage

    def build(
        self,
        collector: ValidationErrorCollector,
        groupby_data: Path | str | None = None,
    ) -> None:
        """Perform build-time validation and create submodels."""
        self.validate_build(collector)
        if collector.has_errors():
            return

        if not (output := self.dataif.get_path(key="output")).exists():
            output.mkdir()

        # Create data subsets
        if self.groupby is not None:
            if groupby_data is None:
                raise ValueError(
                    "groupby_data is required for groupby attribute"
                )
            self.create_subsets(groupby_data)

        # Create parameter sets
        if self.crossby is not None:
            self.create_params()

    def validate_build(
        self, collector: ValidationErrorCollector | None = None
    ) -> None:
        """Perfom build-time validation."""
        self.input.check_missing(collector=collector)

        if self.input_validation:
            for schema in self.input_validation.values():
                if isinstance(schema, Data):
                    schema.validate_metadata(kind="input", collector=collector)

        if self.output_validation:
            for schema in self.output_validation.values():
                if isinstance(schema, Data):
                    schema.validate_metadata(kind="output", collector=collector)

    def validate_run(
        self, collector: ValidationErrorCollector | None = None
    ) -> None:
        """Perfom run-time validation."""
        # TODO: add method arg
        self.input.check_exists(collector=collector)

        if self.input_validation:
            for item_name, schema in self.input_validation.items():
                data_path = self.input.get(item_name)
                if data_path:
                    schema.path = Path(data_path)
                    schema.validate_data(None, collector)
                else:
                    handle_error(
                        self.name,
                        "Input validation",
                        ValueError,
                        f"Input data path for '{item_name}' not found in stage inputs.",
                        collector,
                    )

    def validate_outputs(
        self, collector: ValidationErrorCollector | None = None
    ) -> None:
        """Perform post-run validation of outputs."""
        self.output.check_exists(collector=collector)

        if self.output_validation:
            for item_name, data_spec in self.output_validation.items():
                data_output = self.output.get(item_name)
                if data_output:
                    data_spec.path = Path(data_output.path)
                    data_spec.validate_data(None, collector)
                else:
                    handle_error(
                        self.name,
                        "Output Validation",
                        KeyError,
                        f"Output data '{item_name}' not found after stage execution.",
                        collector,
                    )

    def evaluate(
        self,
        method: Literal["run", "fit", "predict", "collect"],
        subsets: dict[str, Any | list[Any]] | None,
        paramsets: dict[str, Any | list[Any]] | None,
        collect: bool | None,
        backend: Literal["local", "jobmon"],
        cluster: str | None,
        resources: Path | str | dict[str, Any] | None,
        python: Path | str | None,
        collector: ValidationErrorCollector | None = None,
        **kwargs,
    ) -> None:
        """Evaluate stage method."""
        self.validate_run(collector=collector)

        if backend == "jobmon":
            from onemod.backend.jobmon_backend import evaluate_with_jobmon

            if cluster is None:
                raise ValueError("Jobmon backend requires cluster name")
            if resources is None:
                raise ValueError("Jobmon backend requires compute resources")

            evaluate_with_jobmon(
                model=self,
                method=method,
                subsets=subsets,
                paramsets=paramsets,
                collect=collect,
                cluster=cluster,
                resources=resources,
                python=python,
                **kwargs,
            )
        else:
            from onemod.backend.local_backend import evaluate_local

            evaluate_local(
                model=self,
                method=method,
                subsets=subsets,
                paramsets=paramsets,
                collect=collect,
                **kwargs,
            )

        # FIXME: only validate outputs created by method
        # self.validate_outputs(method, collector)

    def run(
        self,
        subsets: dict[str, Any | list[Any]] | None = None,
        paramsets: dict[str, Any | list[Any]] | None = None,
        collect: bool | None = None,
        backend: Literal["local", "jobmon"] = "local",
        cluster: str | None = None,
        resources: Path | str | dict[str, Any] | None = None,
        python: Path | str | None = None,
        **kwargs,
    ) -> None:
        """Run stage.

        Parameters
        ----------
        subsets : dict, optional
            Submodel data subsets to run. If None, run all data subsets.
            Default is None.
        paramsets : dict, optional
            Submodel parameter sets to run. If None, run all parameter
            sets. Default is None.
        collect : bool, optional
            Whether to collect submodel results. If ``subsets`` and
            ``paramsets`` are both None, default is True, otherwise
            default is False.
        backend : {'local', 'jobmon'}, optional
            How to evaluate the method. Default is 'local'.
        **kwargs
            Additional keyword arguments passed to stage method.

        Jobmon Parameters
        -----------------
        cluster : str, optional
            Cluster name. Required if ``backend`` is 'jobmon'.
        resources : Path, str, or dict, optional
            Path to resources file or dictionary of compute resources.
            Required if ``backend`` is 'jobmon'.
        Python : Path or str, optional
            Path to Python environment if ``backend`` is 'jobmon'. If
            None, use sys.executable. Default is None.

        """
        self.evaluate(
            "run",
            subsets,
            paramsets,
            collect,
            backend,
            cluster,
            resources,
            python,
            **kwargs,
        )

    def _run(self, *args, **kwargs) -> None:
        """Run stage submodel.

        Parameters
        ----------
        subset : dict[str, Any]
            If stage uses `groupby` attribute, add `subset` arg when
            implementing `_run`.
        paramset : dict[str, Any]
            If stage uses `crossby` attribute, add `paramset` arg when
            implementing `_run`.

        """
        raise NotImplementedError("Subclasses must implement this method")

    def fit(
        self,
        subsets: dict[str, Any | list[Any]] | None = None,
        paramsets: dict[str, Any | list[Any]] | None = None,
        collect: bool | None = None,
        backend: Literal["local", "jobmon"] = "local",
        cluster: str | None = None,
        resources: Path | str | dict[str, Any] | None = None,
        python: Path | str | None = None,
        **kwargs,
    ) -> None:
        """Fit stage.

        Parameters
        ----------
        subsets : dict, optional
            Submodel data subsets to fit. If None, fit all data subsets.
            Default is None.
        paramsets : dict, optional
            Submodel parameter sets to fit. If None, fit all parameter
            sets. Default is None.
        collect : bool, optional
            Whether to collect submodel results. If ``subsets`` and
            ``paramsets`` are both None, default is True, otherwise
            default is False.
        backend : {'local', 'jobmon'}, optional
            How to evaluate the method. Default is 'local'.
        **kwargs
            Additional keyword arguments passed to stage method.

        Jobmon Parameters
        -----------------
        cluster : str, optional
            Cluster name. Required if ``backend`` is 'jobmon'.
        resources : Path, str, or dict, optional
            Path to resources file or dictionary of compute resources.
            Required if ``backend`` is 'jobmon'.
        python : Path or str, optional
            Path to Python environment if ``backend`` is 'jobmon'. If
            None, use sys.executable. Default is None.

        """
        self.evaluate(
            "fit",
            subsets,
            paramsets,
            collect,
            backend,
            cluster,
            resources,
            python,
            **kwargs,
        )

    def _fit(self, *args, **kwargs) -> None:
        """Fit stage submodel.

        Parameters
        ----------
        subset : dict[str, Any]
            If stage uses `groupby` attribute, add `subset` arg when
            implementing `_fit`.
        paramset : dict[str, Any]
            If stage uses `crossby` attribute, add `paramset` arg when
            implementing `_fit`.

        """
        raise NotImplementedError(
            "Subclasses must implement this method if not skipped"
        )

    def predict(
        self,
        subsets: dict[str, Any | list[Any]] | None = None,
        paramsets: dict[str, Any | list[Any]] | None = None,
        collect: bool | None = None,
        backend: Literal["local", "jobmon"] = "local",
        cluster: str | None = None,
        resources: Path | str | dict[str, Any] | None = None,
        python: Path | str | None = None,
        **kwargs,
    ) -> None:
        """Create stage predictions.

        Parameters
        ----------
        subsets : dict, optional
            Submodel data subsets to create predictions for. If None,
            create predictions for all data subsets. Default is None.
        paramsets : dict, optional
            Submodel parameter sets to create predictions for. If None,
            create predictions for all parameter sets. Default is None.
        collect : bool, optional
            Whether to collect submodel results. If ``subsets`` and
            ``paramsets`` are both None, default is True, otherwise
            default is False.
        backend : {'local', 'jobmon'}, optional
            How to evaluate the method. Default is 'local'.
        **kwargs
            Additional keyword arguments passed to stage method.

        Jobmon Parameters
        -----------------
        cluster : str, optional
            Cluster name. Required if ``backend`` is 'jobmon'.
        resources : Path, str, or dict, optional
            Path to resources file or dictionary of compute resources.
            Required if ``backend`` is 'jobmon'.
        python : Path or str, optional
            Path to Python environment if ``backend`` is 'jobmon'. If
            None, use sys.executable. Default is None.

        """
        self.evaluate(
            "fit",
            subsets,
            paramsets,
            collect,
            backend,
            cluster,
            resources,
            python,
            **kwargs,
        )

    def _predict(self, *args, **kwargs) -> None:
        """Create stage submodel predictions.

        Parameters
        ----------
        subset : dict[str, Any]
            If stage uses `groupby` attribute, add `subset` arg when
            implementing `_predict`.
        paramset : dict[str, Any]
            If stage uses `crossby` attribute, add `paramset` arg when
            implementing `_predict`.

        """
        raise NotImplementedError(
            "Subclasses must implement this method if not skipped"
        )

    def collect(self) -> None:
        """Collect stage submodel results."""
        raise NotImplementedError(
            "Subclasses must implement this method if using submodels"
            " and collect_after not empty"
        )

    def __call__(self, **input: Data | Path | str) -> Output:
        """Define stage dependencies.

        Parameters
        ----------
        **input: Data, Path, or str
            Stage input items.

        Returns
        -------
        Output
            Stage output metadata.

        """
        # TODO: Could add more description, like dataif, input keys
        self.input.update(input)

        for item_name, item_value in self.input.items.items():
            self.dataif.add_path(item_name, item_value.path, exist_ok=True)

        return self.output

    def __repr__(self) -> str:
        stage_str = f"{self.type}(name={self.name}"
        if self.groupby is not None:
            stage_str += f", groupby={self.groupby}"
        if self.crossby is not None:
            stage_str += f", crossby={self.crossby}"
        return stage_str + ")"



"""Functions to run pipelines and stages with Jobmon.

Examples
--------
Compute resources can be passed as a dictionary or a path to a resources
file (e.g., json, toml, yaml).

Required tool resources:

.. code-block:: yaml

    tool_resources:
      {cluster_name}:
        project: {proj_name}
        queue: {queue_name}

To test workflow on a dummy cluster (i.e., run workflow without running
tasks), use:

.. code-block:: yaml

    tool_resources:
      dummy:
        queue: null.q

Optional stage resources can be specified at the stage or stage + method
level:

    task_template_resources:
      {stage_name}:
        {cluster_name}:
            ...
      {stage_name}_{collect}:
        {cluster_name}:
            ...

See Jobmon documentation for additional resources and default values.

"""
# TODO: Optional stage-specific Python environments
# TODO: User-defined max_attempts
# TODO: Could dependencies be method specific?
# TODO: should we check resources format, minimum resources, cluster?

import sys
from pathlib import Path
from typing import Any, Literal

from jobmon.client.api import Tool
from jobmon.client.task import Task
from jobmon.client.task_template import TaskTemplate
from pydantic import validate_call

from onemod.backend.utils import (
    check_input_exists,
    check_method,
    collect_results,
)
from onemod.fsutils.config_loader import ConfigLoader
from onemod.pipeline import Pipeline
from onemod.stage import Stage


@validate_call
def evaluate_with_jobmon(
    model: Pipeline | Stage,
    method: Literal["run", "fit", "predict", "collect"],
    cluster: str,
    resources: Path | str | dict[str, Any],
    python: Path | str | None = None,
    stages: list[str] | None = None,
    subsets: dict[str, Any | list[Any]] | None = None,
    paramsets: dict[str, Any | list[Any]] | None = None,
    collect: bool | None = None,
    **kwargs,
) -> None:
    """Evaluate pipeline or stage method with Jobmon.

    Parameters
    ----------
    model : Pipeline or Stage
        Pipeline or stage instance.
    method : {'run', 'fit', 'predict', 'collect'}
        Name of method to evalaute.
    cluster : str
        Cluster name.
    resources : dict, Path, or str
        Path to resources file or dictionary of compute resources.
    python : Path or str, optional
        Path to Python environment. If None, use sys.executable.
        Default is None.
    **kwargs
        Additional keyword arguments passed to stage methods. If `model`
        is a `Pipeline` instance, use format`stage={arg_name: arg_value}`.

    Pipeline Parameters
    -------------------
    stages : list of str, optional
        Names of stages to evaluate if `model` is a `Pipeline` instance.
        If None, evaluate pipeline stages. Default is None.

    Stage Parameters
    ----------------
    subsets : dict, optional
        Submodel data subsets to evaluate if `model` is a `Stage`
        instance. If None, evaluate all data subsets. Default is None.
    paramsets : dict, optional
        Submodel parameter sets to evaluate if `model` is a `Stage`
        instance. If None, evaluate all parameter sets. Default is None.
    collect : bool, optional
        Whether to collect submodel results if `model` is a `Stage`
        instance. If `subsets` and `paramsets` are both None, default is
        True, otherwise default is False.

    """
    check_method(model, method)
    check_input_exists(model, stages)
    if python is None:
        python = str(sys.executable)

    resources_dict = get_resources(resources)
    tool = get_tool(model.name, method, cluster, resources_dict)
    tasks = get_tasks(
        model,
        method,
        tool,
        resources_dict,
        python,
        stages,
        subsets,
        paramsets,
        collect,
        **kwargs,
    )
    run_workflow(model.name, method, tool, tasks)


def get_resources(resources: Path | str | dict[str, Any]) -> dict[str, Any]:
    """Get dictionary of compute resources.

    Parameters
    ----------
    resources : Path, str, or dict
        Path to resources file or dictionary of compute resources.

    Returns
    -------
    dict
        Dictionary of compute resources.

    """
    if isinstance(resources, (Path, str)):
        config_loader = ConfigLoader()
        return config_loader.load(Path(resources))
    return resources


def get_tool(
    name: str, method: str, cluster: str, resources: dict[str, Any]
) -> Tool:
    """Get Jobmon tool.

    Parameters
    ----------
    name : str
        Pipeline or stage name.
    method : str
        Name of method to evaluate.
    cluster : str
        Cluster name.
    resources : dict
        Dictionary of compute resources.

    Returns
    -------
    Tool
        Jobmon tool.

    """
    tool = Tool(name=f"{name}_{method}")
    tool.set_default_cluster_name(cluster)
    tool.set_default_compute_resources_from_dict(
        cluster, resources["tool_resources"][cluster]
    )
    return tool


def get_tasks(
    model: Pipeline | Stage,
    method: str,
    tool: Tool,
    resources: dict[str, Any],
    python: Path | str,
    stages: list[str] | None,
    subsets: dict[str, Any | list[Any]] | None,
    paramsets: dict[str, Any | list[Any]] | None,
    collect: bool | None,
    **kwargs,
) -> list[Task]:
    """Get Jobmon tasks.

    Parameters
    ----------
    model : Pipeline or Stage
        Pipeline or stage instance.
    method : str
        Name of method to evaluate.
    tool : Tool
        Jobmon tool.
    resources : dict
        Dictionary of compute resources.
    python : Path or str
        Path to Python environment.
    **kwargs
        Additional keyword arguments passed to stage methods.

    Pipeline Parameters
    -------------------
    stages : list of str or None
        Name of stages to evaluate if `model` is a pipeline instance. If
        None, evaluate all pipeline stages.

    Stage Parameters
    ----------------
    subsets : dict or None
        Submodel data subsets to evaluate if `model` is a `Stage`
        instance. If None, evaluate all data subsets.
    paramsets : dict or None
        Submodel parameter sets to evaluate if `model` is a `Stage`
        instance. If None, evaluate all parameter sets.
    collect : bool or None
        Whether to collect submodel results if `model` is a `Stage`
        instance.

    Returns
    -------
    list of Task
        List of Jobmon tasks.

    """
    if isinstance(model, Pipeline):
        return get_pipeline_tasks(
            model, method, tool, resources, python, stages, **kwargs
        )
    return get_stage_tasks(
        model,
        method,
        tool,
        resources,
        python,
        subsets,
        paramsets,
        collect,
        **kwargs,
    )


def get_pipeline_tasks(
    pipeline: Pipeline,
    method: str,
    tool: Tool,
    resources: dict[str, Any],
    python: Path | str,
    stages: list[str] | None,
    **kwargs,
) -> list[Task]:
    """Get pipeline stage tasks.

    Parameters
    ----------
    pipeline : Pipeline
        Pipeline instance.
    method : str
        Name of method to evaluate.
    tool : Tool
        Jobmon tool.
    resources : dict
        Dictionary of compute resources.
    python : Path or str
        Path to Python environment.
    stages : list of str or None
        Name of stages to evaluate. If None, evaluate all pipeline
        stages.
    **kwargs
        Additional keyword arguments passed to stage methods.

    Returns
    -------
    list of Task
        List of pipeline stage tasks.

    """
    tasks = []
    task_dict: dict[str, list[Task]] = {}

    for stage_name in pipeline.get_execution_order(stages):
        stage = pipeline.stages[stage_name]
        if method not in stage.skip:
            upstream_tasks = get_upstream_tasks(
                stage, method, pipeline.stages, task_dict, stages
            )
            task_dict[stage_name] = get_stage_tasks(
                stage,
                method,
                tool,
                resources,
                python,
                upstream_tasks=upstream_tasks,
                **kwargs,
            )
            tasks.extend(task_dict[stage_name])

    return tasks


def get_upstream_tasks(
    stage: Stage,
    method: str,
    stage_dict: dict[str, Stage],
    task_dict: dict[str, list[Task]],
    stages: list[str] | None,
) -> list[Task]:
    """Get upstream tasks for current stage.

    Parameters
    ----------
    stage : Stage
        Current stage instance.
    method : str
        Name of method to evaluate.
    stage_dict : dict
        Dictionary of all pipeline stages.
    task_dict : dict
        Dictionary of all upstream stage tasks.
    stages : list of str or None
        Names of all pipeline stages being evaluated. If None, assume
        all stages are being evaluated.

    Returns
    -------
    list of Task
        Upstream stage tasks for current stage.

    Notes
    -----
    * Only include tasks corresponding to the current stage's
      dependencies that are included in `stages`.
    * If an upstream stage has submodels and `method` is in the
      upstream's `collect_after`, only include the task corresponding to
      the upstream's `collect` method.

    """
    upstream_tasks = []

    for upstream_name in stage.dependencies:
        if stages is not None and upstream_name not in stages:
            # upstream stage is not being evaluated
            continue

        upstream_stage = stage_dict[upstream_name]
        if method not in upstream_stage.skip:
            if (
                upstream_stage.has_submodels
                and method in upstream_stage.collect_after
            ):
                # only include task corresponding to 'collect' method
                upstream_tasks.append(task_dict[upstream_name][-1])
            else:
                upstream_tasks.extend(task_dict[upstream_name])

    return upstream_tasks


def get_stage_tasks(
    stage: Stage,
    method: str,
    tool: Tool,
    resources: dict[str, Any],
    python: Path | str,
    subsets: dict[str, Any | list[Any]] | None = None,
    paramsets: dict[str, Any | list[Any]] | None = None,
    collect: bool | None = None,
    upstream_tasks: list[Task] | None = None,
    **kwargs,
) -> list[Task]:
    """Get stage tasks.

    Parameters
    ----------
    stage : Stage
        Stage instance.
    method : str
        Name of method to evaluate.
    tool : Tool
        Jobmon tool.
    resources : dict
        Dictionary of compute resources.
    python : Path or str
        Path to Python environment.
    subsets : dict, optional
        Submodel data subsets to evaluate. If None, evaluate all data
        subsets. Default is None.
    paramsets : dict, optional
        Submodel parameter sets to evaluate. If None, evaluate all
        parameter sets. Default is None.
    collect : bool, optional
        Whether to collect submodel results if `model` is a `Stage`
        instance. If `subsets` and `paramsets` are both None, default is
        True, otherwise default is False.
    upstream_tasks : list of Task or None, optional
        List of upstream stage tasks. Default is None.
    **kwargs
        Additional keyword arguments passed to stage method.

    Returns
    -------
    list of Task
        List of stage tasks.

    """
    if upstream_tasks is None:
        upstream_tasks = []

    entrypoint = str(Path(python).parent / "onemod")
    config_path = str(stage.dataif.get_path("config"))
    submodel_args = get_submodel_args(stage, method, subsets, paramsets)

    task_template = get_task_template(
        stage.name,
        method,
        tool,
        resources,
        list(submodel_args.keys()),
        **kwargs,
    )

    if submodel_args:
        tasks = task_template.create_tasks(
            name=f"{stage.name}_{method}",
            upstream_tasks=upstream_tasks,
            max_attempts=1,
            entrypoint=entrypoint,
            config=config_path,
            method=method,
            stages=stage.name,
            **{**submodel_args, **kwargs},
        )
    else:
        tasks = [
            task_template.create_task(
                name=f"{stage.name}_{method}",
                upstream_tasks=upstream_tasks,
                max_attempts=1,
                entrypoint=entrypoint,
                config=config_path,
                method=method,
                stages=stage.name,
                **kwargs,
            )
        ]

    if collect_results(stage, method, subsets, paramsets, collect):
        tasks.extend(
            get_stage_tasks(
                stage, "collect", tool, resources, python, upstream_tasks=tasks
            )
        )

    return tasks


def get_submodel_args(
    stage: Stage,
    method: str,
    subsets: dict[str, Any | list[Any]] | None,
    paramsets: dict[str, Any | list[Any]] | None,
) -> dict[str, list[str]]:
    """Get dictionary of subset and/or paramset values.

    If stage has submodels and `method` is not 'collect', additional
    args for 'subsets' and/or 'paramsets' are included in the command
    template.

    Parameters
    ----------
    stage : Stage
        Stage instance.
    method : str
        Method being evaluated.
    subsets : dict or None.
        Submodel data subsets to evaluate. If None, evaluate all data
        subsets.
    paramsets : dict or None
        Submodel parameter sets to evaluate. If None, evaluate all
        parameter sets.

    Returns
    -------
    dict
        Dictionary of subset and/or paramset values as strings.

    """
    submodel_args = {}
    if stage.has_submodels and method != "collect":
        # Get data subsets
        if (filtered_subsets := stage.subsets) is not None:
            if subsets is not None:
                filtered_subsets = stage.get_subset(filtered_subsets, subsets)
            submodel_args["subsets"] = [
                str(subset)
                for subset in filtered_subsets.to_dict(orient="records")
            ]

        # Get parameter sets
        if (filtered_paramsets := stage.paramsets) is not None:
            if paramsets is not None:
                filtered_paramsets = stage.get_subset(
                    filtered_paramsets, paramsets
                )
            submodel_args["paramsets"] = [
                str(paramset)
                for paramset in filtered_paramsets.to_dict(orient="records")
            ]

    return submodel_args


def get_task_template(
    stage_name: str,
    method: str,
    tool: Tool,
    resources: dict[str, Any],
    submodel_args: list[str],
    **kwargs,
) -> TaskTemplate:
    """Get stage task template.

    Parameters
    ----------
    stage_name : str
        Stage name.
    method : str
        Name of method being evaluated.
    tool : Tool
        Jobmon tool.
    resources : dict
        Dictionary of compute resources.
    submodel_args : list of str
        List including 'subsets' and/or 'paramsets'.
    **kwargs
        Additional keyword arguments passed to stage method.

    Returns
    -------
    TaskTemplate
        Stage task template.

    """
    task_template = tool.get_task_template(
        template_name=f"{stage_name}_{method}",
        command_template=get_command_template(method, submodel_args, **kwargs),
        op_args=["entrypoint"],
        task_args=["config", "method", "stages"] + list(kwargs.keys()),
        node_args=submodel_args,
    )

    task_resources = get_task_resources(
        resources, tool.default_cluster_name, stage_name, method
    )
    if task_resources:
        task_template.set_default_compute_resources_from_dict(
            tool.default_cluster_name, task_resources
        )

    return task_template


def get_command_template(
    method: str, submodel_args: list[str], **kwargs
) -> str:
    """Get stage command template.

    All stages methods are called via `onemod.main.evaluate()`. If stage
    has submodels and `method` is not 'collect', additional args for
    'subsets' and/or 'paramsets' are included in the command template.

    Parameters
    ----------
    method : str
        Name of method being evaluated.
    submodel_args : list of str
        List including 'subsets' and/or 'paramsets'.
    **kwargs
        Additional keyword arguments passed to stage method.

    Returns
    -------
    str
        Stage command template.

    # TODO: collapse into one list

    """
    command_template = (
        "{entrypoint} --config {config} --method {method} --stages {stages}"
    )

    if method != "collect":
        for key, value in kwargs.items():
            if isinstance(value, (dict, list, set, tuple)):
                command_template += f" --{key} '{{{key}}}'"
            else:
                command_template += f" --{key} {{{key}}}"

        for arg in submodel_args:
            command_template += f" --{arg} '{{{arg}}}'"

    return command_template


def get_task_resources(
    resources: dict[str, Any], cluster: str, stage_name: str, method: str
) -> dict[str, Any]:
    """Get task-specific resources.

    Parameters
    ----------
    resources : dict
        Dictionary of compute resources.
    cluster : str
        Cluster name.
    stage_name : str
        Stage name.
    method : str
        Name of method being evaluated.

    Returns
    -------
    dict
        Task-specific resources.

    """
    task_resources = resources.get("task_template_resources", {})
    stage_resources = task_resources.get(stage_name, {})
    method_resources = task_resources.get(f"{stage_name}_{method}", {})
    return {
        **stage_resources.get(cluster, {}),
        **method_resources.get(cluster, {}),
    }


def run_workflow(name: str, method: str, tool: Tool, tasks: list[Task]) -> None:
    """Create and run workflow.

    Parameters
    ----------
    name : str
        Pipeline or stage name.
    method : str
        Name of method being evaluated.
    tool : Tool
        Jobmon tool.
    tasks : list of Task
        List of stage tasks.

    """
    workflow = tool.create_workflow(name=f"{name}_{method}")
    workflow.add_tasks(tasks)
    workflow.bind()
    print(f"Starting workflow {workflow.workflow_id}")
    status = workflow.run()
    if status != "D":
        raise ValueError(f"Workflow {workflow.workflow_id} failed")
    else:
        print(f"Workflow {workflow.workflow_id} finished")



import json
from pathlib import Path
from typing import Any, TypeVar

import yaml
from pydantic import BaseModel
from yaml import Node, SafeDumper

T = TypeVar("T", bound=BaseModel)


def _yaml_path_representer(dumper: SafeDumper, data: Path) -> Node:
    return dumper.represent_str(str(data))


def _yaml_set_representer(dumper: SafeDumper, data: set) -> Node:
    return dumper.represent_list(list(data))


yaml.add_multi_representer(Path, _yaml_path_representer, SafeDumper)
yaml.add_representer(set, _yaml_set_representer, SafeDumper)


def deserialize(filepath: Path | str) -> dict:
    """Load a configuration from a YAML or JSON file."""
    file_format = _get_file_format(filepath)
    with open(filepath, "r") as file:
        if file_format == "json":
            return json.load(file)
        elif file_format == "yaml":
            return yaml.safe_load(file)
        else:
            raise ValueError(f"Unsupported file format: {file_format}")


def serialize(obj: T | dict | list[T | dict], filepath: Path | str) -> None:
    """Save a Pydantic model, dict, or list of these to a YAML or JSON file."""
    file_format = _get_file_format(filepath)
    data = _normalize(obj)

    with open(filepath, "w") as file:
        if file_format == "json":
            json.dump(data, file, indent=4, cls=_CustomJSONEncoder)
        elif file_format == "yaml":
            yaml.safe_dump(data, file, default_flow_style=False)


class _CustomJSONEncoder(json.JSONEncoder):
    """Custom JSON Encoder to handle Path, set, and other unsupported types."""

    def default(self, obj: Any) -> Any:
        if isinstance(obj, Path):
            return str(obj)
        elif isinstance(obj, set):
            return list(obj)
        return super().default(obj)


def _get_file_format(filepath: Path | str) -> str:
    """Detect the file format based on the file extension."""
    file_extension = Path(filepath).suffix.lower()
    if file_extension in {".json"}:
        return "json"
    elif file_extension in {".yaml", ".yml"}:
        return "yaml"
    else:
        raise ValueError(
            f"Unsupported file format: {file_extension}. Supported formats are: .json, .yaml, .yml"
        )


def _normalize(obj: Any) -> Any:
    """Convert Pydantic models and sets to serializable forms."""
    if isinstance(obj, BaseModel):
        return obj.model_dump()
    elif isinstance(obj, dict):
        return {k: _normalize(v) for k, v in obj.items()}
    elif isinstance(obj, list):
        return [_normalize(item) for item in obj]
    elif isinstance(obj, set):
        return list(obj)
    elif isinstance(obj, Path):
        return str(obj)
    else:
        return obj



from contextlib import contextmanager
from typing import Any, Generator

from pydantic import BaseModel


class ValidationErrorReport(BaseModel):
    stage: str | None
    error_category: str
    message: str
    details: dict[str, Any] | None = None


class ValidationErrorCollector(BaseModel):
    errors: list[ValidationErrorReport] = []

    def add_error(
        self,
        stage: str | None,
        error_category: str,
        message: str,
        details: dict | None = None,
    ) -> None:
        error_report = ValidationErrorReport(
            stage=stage,
            error_category=error_category,
            message=message,
            details=details,
        )
        self.errors.append(error_report)

    def has_errors(self) -> bool:
        return len(self.errors) > 0

    def get_errors(self) -> list[ValidationErrorReport]:
        return self.errors

    def raise_errors(self) -> None:
        if self.has_errors():
            raise ValidationException(self.errors)


class ValidationException(Exception):
    def __init__(self, errors: list[ValidationErrorReport]):
        self.errors = errors
        super().__init__(self._format_errors())

    def _format_errors(self) -> str:
        messages = [
            f"Stage '{error.stage}' - {error.error_category}: {error.message}"
            for error in self.errors
        ]
        return "\n".join(messages)


@contextmanager
def validation_context() -> Generator[ValidationErrorCollector, None, None]:
    """Context manager for managing validation error collection."""
    validation_collector = ValidationErrorCollector()
    try:
        yield validation_collector
    finally:
        pass


def handle_error(
    stage: str | None,
    error_category: str,
    error_type: type[Exception],
    message: str,
    collector: ValidationErrorCollector | None = None,
    details: dict | None = None,
) -> None:
    """Handle an error by either raising it or adding it to the validation collector."""
    if collector:
        collector.add_error(stage, error_category, message, details)
    else:
        raise error_type(message)



from collections.abc import Collection
from typing import Any, Callable

from polars import Series


def bounds(
    ge: float | None = None, le: float | None = None
) -> Callable[[Series], None]:
    """
    Returns a function that checks if all values in the column are within the specified bounds.

    Parameters
    ----------
    ge (float): Minimum allowed value (inclusive)
    le (float): Maximum allowed value (inclusive)

    Examples
    --------
    >>> use_validation = bounds(ge=0, le=100)
    >>> use_validation(pl.Series([1, 2, 3, 4, 5]))  # No error
    >>> use_validation(
    ...     pl.Series([-1, 2, 3, 4, 5])
    ... )  # ValueError: All values must be greater than or equal to 0.

    Returns
    -------
        callable: Function that validates the column values
    """

    def use_validation(column: Series) -> None:
        if ge is not None:
            if not column.ge(ge).all():
                raise ValueError(
                    f"All values must be greater than or equal to {ge}."
                )

        if le is not None:
            if not column.le(le).all():
                raise ValueError(
                    f"All values must be less than or equal to {le}."
                )

    return use_validation


def is_in(other: Collection[Any]) -> Callable[[Series], None]:
    """
    Returns a function that checks if all values in the column are within the specified collection.

    Parameters
    ----------
    other (Collection[Any]): Collection of values

    Examples
    --------
    >>> use_validation = is_in(["a", "b", "c"])
    >>> use_validation(pl.Series(["a", "b"]))  # No error
    >>> use_validation(
    ...     pl.Series(["a", "d"])
    ... )  # ValueError: All values must be in ['a', 'b', 'c'].

    Returns
    -------
        callable: Function that validates the column values
    """

    def use_validation(column: Series) -> None:
        if not column.is_in(other).all():
            raise ValueError(f"All values must be in {other}.")

    return use_validation


def no_inf() -> Callable[[Series], None]:
    """
    Returns a function that checks that there are no infinite(inf, -inf) values in the column.

    Examples
    --------
    >>> use_validation = no_inf()
    >>> use_validation(pl.Series([1, 2, 3, 4, 5]))  # No error
    >>> use_validation(
    ...     pl.Series([1, 2, float("inf"), 4, 5])
    ... )  # ValueError: All values must be finite.

    Returns
    -------
        callable: Function that validates the column values
    """

    def use_validation(column: Series) -> None:
        if column.is_in([float("inf"), float("-inf")]).any():
            raise ValueError("All values must be finite.")

    return use_validation
